// lzw_dictionary.cpp: implementation of the dictionary class.
// Created by: Luria Israel
//  This project implements the use of a dynamic LZW compression.
//  This code is free for use, and can be inserted into any MFC project.
//
// Modified by Pete Kirkham. Heavily.

#include "lzw/lzw_compression.h"

#include "utils/string_utils.h"

#define null 0

#ifndef posix_memalign
#define posix_memalign(pv, a, s) ((*(pv))=malloc(s))
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
// (why do MFC programs always have such inane comments?)
//////////////////////////////////////////////////////////////////////
namespace lzw
{
    inline unsigned int element (unsigned int prefix, char letter)
    {
        return (prefix << 8) | (letter & 0xff);
    }

    inline unsigned int storage_ref (unsigned int data, unsigned int mask)
    {
        if (data&0xff000000)
        {
            std::cout << "data outside of range." << std::endl;
            exit(-1);
        }

        if (data&1)
        {
            std::cout << "data index not even." << std::endl;
            exit(-1);
        }

        if (mask&0xffffff00)
        {
            std::cout << "mask excedes 8 bit limit." << std::endl;
            exit(-1);
        }

        return (data << 7) | mask;
    }

    dictionary::dictionary (unsigned int heap_size_in_bytes, bool expanded) : _heap_size(heap_size_in_bytes/4), _max_nodes(heap_size_in_bytes/8), _expanded(expanded)
    {
        if (sizeof(int) != 4)
        {
            std::cerr << "int not 32 bits " << std::endl;
            exit(-1);
        }

        // allow 64MB with 50% nodes
        if (_max_nodes == 0x800000)
        {
            _max_nodes = 0x400000;
        }

        if (_max_nodes >= 0x800000)
        {
            std::cerr << "cannot support heaps of more than " << 0x800000 << " nodes.";
            exit(-1);
        }

        if (heap_size_in_bytes&(heap_size_in_bytes-1))
        {
            std::cerr << "heap size should be a power of 2.";
            exit(-1);
        }

        if (heap_size_in_bytes < 1024)
        {
            std::cerr << "heap size should be at least 1K.";
            exit(-1);
        }

        posix_memalign(reinterpret_cast<void**>(&_node_data), 16, heap_size_in_bytes);
    }

    dictionary::~dictionary ()
    {
        free(_node_data);
    }

    // resets the dictionary
    void dictionary::reset ()
    {
        for (size_t i(0); i < _heap_size; ++i)
        {
            _node_data[i] = 0xffffffff;
        }

        _node_count = 0;
        _last_heap_index = _heap_size - 1;
        _full = false;

        for (size_t i(0); i <= 255; ++i)
        {
            add_entry(0, char(i));
        }
    }

    // returns offset into data for a free block of size 'size' between top and bottom, starting from top
    // returns -1 if none found.
    size_t find_free_heap (size_t size, unsigned int* data, size_t& _top, size_t bottom, size_t max_index)
    {
        size_t top(_top);

        //~ std::cout << "find_free_heap(" << size << ',' << to_hex(data) << ',' << top << ',' << bottom << ')' << std::endl;

        bottom += size;

        bool    restarted(false);

        for (;;)
        {
            while (data[top] != -1)
            {
                --top;

                if  (top <= bottom)
                {
                    if (restarted)
                    {
                        //~ std::cout << "find_free_heap => -1" << std::endl;
                        return (unsigned int)-1;
                    }
                    else
                    {
                        top = max_index;
                        restarted = true;
                    }
                }
            }

            bool valid = true;

            if (!(top & 1))
            {
                --top;
            }

            for (size_t i(0); i < size; ++i)
            {
                if (data[top - i] != -1)
                {
                    valid = false;
                    top -= i;
                    break;
                }
            }

            if (valid)
            {
                // zero memory to show in use
                for (size_t i(0); i < size; ++i)
                {
                    data[top - i] = 0;
                }

              //  std::cout << "find_free_heap => " << top - size << std::endl;

                return _top = top - size + 1;
            }
        }
    }

    //#define watch_prefix 199
    // adds the entry, returns its index
    void dictionary::add_entry   (unsigned int prefix, char letter)
    {
        //~ if (prefix)
        //~ std::cout << "add_entry(" << prefix << ',' << (letter&0xff) << ')' << " @ " << _node_count << std::endl;

        if (_full || (_node_count >= _max_nodes) || (_node_data[(_node_count<<1)] != 0xffffffff) || (_node_data[(_node_count<<1)|1] != 0xffffffff))
        {
            _full = true;
            return;
        }

#ifdef watch_prefix
        if (prefix == watch_prefix)
        {
            std::cerr << "add_entry(" << prefix << "," << int(letter&0xff) << ")" << " node_count: " << _node_count << std::endl;
            dump_node(prefix);
        }
#endif
        unsigned int elem(element(prefix, letter));
        unsigned int id(_node_count);

        if (prefix)
        {
            if (prefix > _node_count)
            {
                std::cerr << "invalid prefix: " << prefix << " " << to_binary(size_t(prefix)) << std::endl;
                exit(-1);
                return;
            }

            // find prefix's child storage
            unsigned int& storage(_node_data[(prefix << 1)|1]);

            if (storage == 0) // no children
            {
                storage = element(id, letter) | 0x80000000;
            }
            else
            {
                size_t mask;
                size_t data;

                if (storage & 0x80000000) // single child
                {
                    storage &= 0x7fffffff;
                    data = (prefix << 1)|1;
                    mask = 0;
                }
                else
                {
                    data = (storage & 0x7fffff00) >> 7;
                    mask =  storage & 0x000000ff;
                }

                size_t allocated(mask+1);
                size_t original_allocated(allocated+1);

                if (allocated == 1) //  expand to list of 8 elements
                {
                    size_t new_data(find_free_heap(8, _node_data, _last_heap_index, _node_count*2, _heap_size));

                    if (new_data == -1) return;

                    _node_data[new_data]   = _node_data[data];
                    _node_data[new_data+1] = element(id, letter);
                    _node_data[(id<<1)]    = element(prefix, letter);
                    _node_data[(id<<1)|1]  = 0;
                    ++_node_count;

                    storage = storage_ref(new_data, 7);

                    return;
                }

                if (allocated == 8)
                {
                    for (size_t i(0); i < 8; ++i)
                    {
                        if (_node_data[data + i] == 0)
                        {
                            _node_data[data + i]   = element(id, letter);
                            _node_data[(id<<1)]    = element(prefix, letter);
                            _node_data[(id<<1)|1]  = 0;
                            ++_node_count;

                            return;
                        }
                    }
                }

                // test whether data needs expanding
                int node_data(_node_data[data + (letter&mask)]);

                if (node_data)
                {
                    // double allocated until collision freed
                    char colliding_letter(node_data&0xff);

                    if (colliding_letter == letter)
                    {
                        std::cerr << "re-inserting value " << prefix << "," << (letter&0xff) << std::endl;
                        return;
                    }
                    else
                    {
                        unsigned int new_allocated(allocated<<1);
                        unsigned int new_mask;

                        for (;;)
                        {
                            new_mask = new_allocated-1;

                            if ((colliding_letter&new_mask) != (letter&new_mask)) // collision freed
                            {
                                break;
                            }

                            new_allocated <<= 1;
                        }

                        size_t new_data(find_free_heap(new_allocated, _node_data, _last_heap_index, _node_count*2, _heap_size));

                        if (new_data == -1)
                        {
                            return;
                        }

                        // copy to new_data, marking old data as free
                        for (size_t i(0); i < allocated; ++i)
                        {
                            unsigned int node_value(_node_data[i+data]);

                            if (node_value)
                            {
                                _node_data[new_data + (node_value&new_mask)] = node_value;
                            }

                            _node_data[i+data] = 0xffffffff;
                        }

                        mask        = new_mask;
                        allocated   = new_allocated;
                        data        = new_data;

                        storage     = storage_ref(data, mask);
                    }
                }

                // set new data
                _node_data[data + (letter&mask)] = element(id, letter);
            }
        }

        // add element to node list:
        _node_data[id << 1] = elem;
        _node_data[(id << 1)|1] = 0;

        ++_node_count;

    }

    // finds an entry index
    unsigned int dictionary::get_entry (unsigned int prefix, char letter)
    {
        //~ std::cerr << "get_entry(" << prefix << "," << int(letter) << ")" << std::endl;

        if ((prefix == 0) && ((letter&0x80) == 0))
        {
            return (unsigned int)(letter);
        }

        if (prefix >= _node_count)
        {
            std::cerr << "asking for a prefix that doesn't exist: " << prefix << " >= " << _node_count << std::endl;
            //dump();
            return (unsigned int)-1;
        }

        // find prefix's child storage
        unsigned int node_data(_node_data[(prefix << 1)|1]);

        if (node_data & 0x80000000) // single child
        {
            node_data &= 0x7fffffff;
        }
        else // multiple children
        {
            size_t mask(node_data&0xff);
            size_t data((node_data& 0x7fffff00) >> 7);

            if (mask == 7) // linear scan
            {
                for (size_t i(0); i < 8; ++i)
                {
                    node_data = _node_data[data + i];

                    if (node_data == 0)
                    {
                        return (unsigned int)-1;
                    }

                    if (((node_data^letter) & 0xff) == 0)
                    {
                        break;
                    }
                }
            }
            else
            {
                node_data = _node_data[data + (letter&mask)];
            }
        }

        if ((node_data^letter) & 0xff)
        {
            return (unsigned int)-1;
        }
        else
        {
            unsigned int id(node_data >> 8);

            if (id >= _node_count)
            {
                std::cerr << "invalid id returned from get_entry(" << prefix << ',' << (letter&0xff) << "), " << id << " >= " << _node_count << std::endl;
                dump_node(prefix);
            exit(-1);
                }

            return id;
        }
    }

    unsigned int dictionary::get_prefix (unsigned int code)
    {
        return (_node_data[code<<1] >> 8);
    }

    unsigned int dictionary::get_letter (unsigned int code)
    {
        return (_node_data[code<<1] & 0xff);
    }

    bool dictionary::code_exists (unsigned int code)
    {
        return code < _node_count;
    }

    unsigned int dictionary::size()
    {
        return _expanded ? (_max_nodes-1) : _node_count;
    }

    unsigned int dictionary::write_bytes (char* buffer, size_t& index, size_t limit, unsigned int code)
    {
        //~ std::cerr << "write_bytes (" << to_hex(buffer) << "," << index << "," << limit << "," << code << ")" << std::endl;

        if (code >= _node_count)
        {
            std::cerr << "invalid code " << code << " >= " << _node_count << std::endl;
//            exit(-1);
            return (unsigned int)-1;
        }

        unsigned int elem(_node_data[code<<1]);
        unsigned int prefix(elem >> 8);
        unsigned int prefix_result;

        if (prefix)
        {
            prefix_result = write_bytes(buffer, index, limit, prefix);

            if (prefix_result == -1)
            {
                dump_node(prefix);
                return (unsigned int)-1;
            }
        }

        if (index < limit)
        {
            buffer[index] = char(elem);

            //~ std::cout << code << " -> (" << to_hex(char(elem)) << ')' << std::endl;

            ++index;
        }

        return prefix ? prefix_result : (elem & 0xff);
    }

    void    dictionary::dump ()
    {
        for (size_t index(0); index < _node_count; ++index)
        {
            char    buf[1024];
            size_t  written(0);

            write_bytes(buf, written, 1024, index);

            buf[written] = 0;

            std::cerr << to_trimmed_binary(index) << " => " << to_binary(buf) << std::endl;
        }
    }

    void    dictionary::dump_node (size_t index)
    {
        unsigned int d0(_node_data[index << 1]);
        unsigned int d1(_node_data[(index << 1)+1]);

        std::cerr << index << " =>  </" << (d0 >> 8) << ',' << (d0 & 0xff) << ',' << (d1 >> 8) << ',' << (d1 & 0xff) << "/> [";

        if (d1 != 0) // no children
        {
            if (d1 & 0x80000000) // single child
            {
                std::cerr << "</" << ((d1 & 0x7fffff00) >> 8) << ',' << (d1 & 0xff) << "/>";
            }
            else // multiple children
            {
                size_t mask(d1&0xff);
                size_t size(mask + 1);
                size_t data((d1& 0x7fffff00) >> 7);

                for (size_t j(0); j < size; ++j) // 2 children not using mask
                {
                    if (j > 0) std::cerr << ',';
                    std::cerr << "</" << (_node_data[data+j] >>8)   << ',' <<  (_node_data[data+j] & 0xff) << "/>";
                }
            }
        }

        std::cerr << ']' << std::endl;
    }

    void    dictionary::metrics ()
    {
        size_t sizes[256];
        size_t allocations[256];
        size_t list_total(0);
        size_t list_use(0);

        for (size_t index(0); index < 256; ++index)
        {
            sizes[index] = 0;
            allocations[index] = 0;
        }

        for (size_t index(0); index < _node_count; ++index)
        {
            unsigned int d0(_node_data[index << 1]);
            unsigned int d1(_node_data[(index << 1)+1]);

            size_t  size(0);
            size_t  allocated(0);

            if (d1 == 0) // no children
            {
                size = 0;
                allocated = 0;
            }
            else if (d1 & 0x80000000) // single child
            {
                size = 1;
                allocated = 1; // allocated is extra from the top, not internal to node
            }
            else // multiple children
            {
                size_t mask(d1&0xff);
                allocated = (mask + 1);
                size_t data((d1&0x7fffff00) >> 7);

                for (size_t j(0); j < allocated; ++j)
                {
                    if (_node_data[data+j]) ++size;
                }

                // simulate effect of linear probe on packing
                //~ if (size <= 8)
                //~ {
                    //~ while (allocated > size*2) allocated /= 2;
                //~ }

                list_total += allocated;
                list_use   += size;
            }

            ++sizes[size];
            ++allocations[allocated];
        }

        std::cout << "maximum nodes: \t" << _max_nodes << std::endl;
        std::cout << "total nodes: \t" << _node_count << std::endl;
        std::cout << "leaf nodes:  \t" << sizes[0]  << " (" << (100*sizes[0])/_node_count << "%)" << std::endl;

        size_t  or_fewer(sizes[0]);

        for (size_t index(1); index < 256; ++index)
        {
            if (sizes[index])
            {
                or_fewer += sizes[index];
                std::cout << "nodes with " << index << " children: " << sizes[index];
                std::cout << " (or fewer: " << or_fewer << ", " << (100*or_fewer)/_node_count << "%)" << std::endl;
            }
        }

        or_fewer = 0;

        for (size_t index(0); index < 256; ++index)
        {
            if (allocations[index])
            {
                or_fewer += allocations[index];
                std::cout << "nodes with storage for " << index << " children: " << allocations[index];
                std::cout << " (or fewer: " << or_fewer << ", " << (100*or_fewer)/_node_count << "%)" << std::endl;
            }
        }

        std::cout << "table is " << (_full ? "full" : "not full") << std::endl;
        std::cout << "storage for nodes:         " << ((100*(2*_node_count))/_heap_size) << '%' << std::endl;
        std::cout << "storage for child lists:   " << ((100*(_heap_size-2*_node_count))/_heap_size) << '%' << std::endl;
        std::cout << "total slots in lists:      " << list_total << std::endl;
        std::cout << "non-empty slots in lists:  " << list_use << std::endl;
        std::cout << "list internal efficiency:  " << (100*list_use)/(list_total) << '%' << " (ratio of slots in use to slots allocated)" << std::endl;
        std::cout << "list fragmentation:        " << (100*list_total)/(_heap_size-2*_node_count) << '%' << " (ratio of slots allocated to heap used)" << std::endl;
        std::cout << "list total efficiency:     " << (100*list_use)/(_heap_size-2*_node_count) << '%' << " (ratio of slots in use to heap used)" << std::endl;

        //~ for (size_t i(0); i < _heap_size; i+= 1024)
        //~ {
            //~ bool in_use(false);

            //~ for (size_t j(0); j < 1024; j+=8)
            //~ {
                //~ for (size_t k(0); k < 8; ++k)
                //~ {
                    //~ if (_node_data[i+j+k] != -1)
                    //~ {
                        //~ in_use = true;
                        //~ break;
                    //~ }
                //~ }

                //~ std::cout << in_use ? '*' : ' ';
            //~ }

            //~ std::cout << std::endl;
        //~ }
    }

#define chunk_size (16*1024)

    bool    dictionary::load_from_file  (int file)
    {
        int buf[chunk_size];

        int nread =  ::read(file, buf, chunk_size);

        if (nread > 0)
        {
            std::cerr << "read failed." << std::endl;
            return false;
        }


    }

    bool    dictionary::save_to_file    (int file)
    {

    }
}
